from AccessControl import ClassSecurityInfo
from Acquisition import Implicit
from Acquisition import aq_base, aq_acquire
from Globals import InitializeClass
from OFS.ObjectManager import ObjectManager
from Products.CMFCore  import CMFCorePermissions
from Products.CMFCore.utils import getToolByName
from ZPublisher.HTTPRequest import FileUpload
from debug import log, log_exc
from types import FileType
import operator

from BaseUnit import BaseUnit
from Field  import ObjectField, FieldList
from Form import StringInfo, IdInfo
from utils import DisplayList

content_type = FieldList((
    ObjectField('id',
                required=1,
                mode="rw",
                accessor="getId",
                mutator="setId",
                default=None,
                form_info=IdInfo(label="Name",
                                 description="""Enter the Name
                                 of the document. This is the part
                                 that will show up as part of the 
                                 web address. This should be
                                 short, descriptive and
                                 preferrably not 
                                 contain spaces, underscores or
                                 mixed case. Example: 
                                 'expense-report-2002'""")),
    
    ObjectField('title',
                required=1,
                searchable=1,
		default='',
                form_info=StringInfo(label="Title",
                                     description="""Enter a title
                                     for this content. 
                                     This will appear in the
                                     management screen and
                                     will appear as the text in a
                                     link
                                     to this item."""),),
    ))

class BaseObject(Implicit):
    security = ClassSecurityInfo()

    type = content_type
    installMode = ['type', 'actions', 'navigation', 'validation']
    
    def __init__(self, oid, **kwargs):
        self.initalizeLayers()
        self.setDefaults()
        self.update(**kwargs)
        self.id = oid

    def initalizeLayers(self):
        self.type.initalizeLayers(self)

    security.declarePublic("getId")
    def getId(self):
        """get the objects id"""
        return self.id

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'setId')
    def setId(self, value):
        if value != self.getId():
            try:
                context.manage_renameObjects((self.id,), (value,), self.REQUEST)
            except:
                if hasattr(self, 'aq_parent'):
                    self.aq_parent.manage_renameObjects((self.id,), (value,), self.REQUEST)
        self.id = value
    
    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'getField')
    def getField(self, key):
        return self.type.get(key)

    security.declareProtected(CMFCorePermissions.View, 'getDefault')
    def getDefault(self, field):
        """return the default from the type info"""
        field = self.getField(field)
        return field.default
    
    security.declareProtected(CMFCorePermissions.View, 'isBinary')
    def isBinary(self, key):
        element = getattr(self, key, None)
        if element and hasattr(element, 'isBinary'):
            return element.isBinary()
        return 0

    security.declareProtected(CMFCorePermissions.View, 'getContentType')
    def getContentType(self, key):
        value = 'text/plain' #this should maybe be octet stream or something?
        element = getattr(self, key, None)
        if element and hasattr(element, 'getContentType'):
            return element.getContentType()
        return value

    security.declareProtected(CMFCorePermissions.View, 'Vocabulary')
    def Vocabulary(self, key):
        field = self.getField(key)
        if field:
            return field.Vocabulary(self), field.enforceVocabulary
        return DisplayList(), 0
    

    security.declareProtected(CMFCorePermissions.View, 'getUnit')
    def getUnit(self, key):
        element = getattr(self, key, None)
        if hasattr(element, 'isUnit'):
            return element
        return None
        
    security.declareProtected(CMFCorePermissions.View, 'get')
    def get(self, key, **kwargs):
        """return editable version of content"""
        return self.type[key].get(self, **kwargs)

    def __getitem__(self, key):
        """play nice with externaleditor again"""
        ## Also play nice with aq again... doh!
        if key not in self.type.keys() and key[:1] != "_": #XXX 2.2
            return getattr(self, key)
        return self.get(key)

    def set(self, key, value, **kw):
        self.type[key].set(self, value, **kw)
        
    def edit(self, **kwargs):
        self.update(**kwargs)

    def setDefaults(self):
        self.type.setDefaults(self)
            
    def update(self, **kwargs):
        self.type.updateAll(self, **kwargs)
        self._p_changed = 1
        self.reindexObject()

        
    def validate_field(self, name, value, errors):
        """
        write a method: validate_foo(new_value) -> "error" or None
        If there is a validate method defined for a given field invoke it by name
        name -- the name to register errors under
        value -- the proposed new value
        errors -- dict to record errors in
        """
        methodName = "validate_%s" % name

        base = aq_base(self)
        if hasattr(base, methodName):
            method = getattr(base, methodName)
            result = method(value)
            if result is not None:
                errors[name] = result


    ##Pre/post validate hooks that will need to write errors
    ##into the errors dict directly using errors[fieldname] = ""
    security.declareProtected(CMFCorePermissions.View, 'pre_validate')            
    def pre_validate(self, REQUEST, errors):
        pass

    security.declareProtected(CMFCorePermissions.View, 'post_validate')            
    def post_validate(self, REQUEST, errors):
        pass
    
            
    security.declareProtected(CMFCorePermissions.View, 'validate')            
    def validate(self, REQUEST=None, errors=None):
        if errors is None:
            errors = {}
        self.pre_validate(REQUEST, errors)
        if errors:
            return errors
        
        self.type.validate(self, REQUEST=REQUEST, errors=errors)
        self.post_validate(REQUEST, errors)

        return errors
    
    security.declareProtected(CMFCorePermissions.View, 'SearchableText')
    def SearchableText(self):
        """full indexable text"""
        data = []
        for field in self.type.fields():
            if field.searchable != 1: continue
            try:
                method = getattr(self, field.accessor) ##Editable messes with binary types
                data.append(method())
            except:
                pass

        data = [str(d) for d in data if d is not None]
        data = ' '.join(data)
        return data

    
    security.declareProtected(CMFCorePermissions.View, 'get_size' )
    def get_size( self ):
        """ Used for FTP and apparently the ZMI now too """
        size = 0
        for name in self.type.keys():
            field = getattr(self, name, None)
            if hasattr(field, "isUnit"):
                size += field.get_size()
            else:
                try:
                    size += len(field)
                except:
                    pass
                
        return size


    
InitializeClass(BaseObject)

