from __future__ import nested_scopes
from AccessControl import ClassSecurityInfo
from types import ListType, TupleType
from UserDict import UserDict
from Products.CMFCore  import CMFCorePermissions
from Globals import InitializeClass
from Form import StringInfo, DateTimeInfo
from utils import capitalize, DisplayList
from debug import log, log_exc
from ZPublisher.HTTPRequest import FileUpload
from BaseUnit import BaseUnit
from types import StringType, FileType
from Storage import AttributeStorage, MetadataStorage
from DateTime import DateTime
from Layer import ILayerContainer, ILayerRuntime


#Reusable instance for MetadataFieldList
MDS = MetadataStorage()

# Used in fields() method
def index_sort(a, b): return  a.index - b.index


class FieldList(UserDict):
    """Manage a list of fields and run methods over them"""

    __implements__ = ILayerRuntime
    
    security = ClassSecurityInfo()
    security.declareObjectPublic()
    security.setDefaultAccess("allow")

    def __init__(self, *args, **kwargs):
        UserDict.__init__(self)
        self.order_fields = None
        self.index = 0
        if len(args):
            if type(args[0]) in [ListType, TupleType]:
                for field in args[0]:
                    self.addField(field)
            else:
                self.addField(args[0])
                
        self.props = kwargs

    def __add__(self, other):
        c = FieldList()
        #We can't use update and keep the order so we do it manually
        for field in self.fields():
            c.addField(field)
        for field in other.fields():
            c.addField(field)
        return c


    def fields(self):
        """list all the fields in order"""
        if self.order_fields is None:
            f = self.values()
            f.sort(index_sort)
            self.order_fields = f

        return self.order_fields

    security.declarePublic('fields')
    def filterFields(self, *args, **kwargs):
        """Args is a list of callable conditions
        arg(field) -> 0, omit field, 1 keep field
        kwargs is a list of attribute -> valid value mappings
        if the field doesn't match the rhs then its ommited
        """

        results = []
        fields = self.fields()

        for field in fields:
            skip = 0
            for arg in args:
                if arg(field) == 0:
                    skip = 1 
                    break
                
            for k, v in kwargs.items():
                if hasattr(field, k):
                    fv = getattr(field, k)
                    if fv != v:
                        skip = 1
                        break

            if not skip:
                results.append(field)

        return results
    
        

    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'edit')
    def edit(self, instance, name, value):
        if self.allow(name):
            instance[name] = value

    def setDefaults(self, instance):
        """Only call during object initalization. Sets fields to
        schema defaults
        """
        for field in self.values():
            field.set(instance, field.default)
                
    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'updateAll')
    def updateAll(self, instance, **kwargs):
        keys = kwargs.keys()
        for field in self.values():
            if 'w' not in field.mode:
                log("tried to update %s:%s which is not writeable" % \
                    (instance.portal_type, field.name))
                continue
            
            method = getattr(instance, field.mutator, None)
            if not method:
                log("No method %s on %s" % (field.mutator, instance))
                continue

            if field.name in keys:
                method(kwargs[field.name])
            
    def addField(self, field):
        if isinstance(field, Field):
            self[field.name] = field
            field.index = self.index
            self.index +=1

    security.declarePublic("allow")
    def allow(self, key):
        """Allow update to keys of this name (must be a valid field)"""
        return self.get(key) != None

    security.declarePublic('validate')
    def validate(self, instance, REQUEST=None, errors=None):
        """Validate the state of the entire object"""
        for name, field in self.items():
            if errors and errors.has_key(name): continue
            error = 0
            value = None
            if REQUEST:
                form = REQUEST.form
                for postfix in ['_file', '']:
                    value = form.get("%s%s" % (name, postfix), None)
                    if type(value) != type(''):
                        if isinstance(value, FileUpload):
                            if value.filename == '': continue
                            else: break
                        else:
                            #Do other types need special handling here
                            pass
                        
                    if value is not None and value != '': break

            #REQUIRED CHECK
            if field.required == 1:
                if not value or value == "":
                    ## The only time a field would not be resubmitted with the form is if
                    ## was a file object from a previous edit. That will not come back. 
                    ## We have to check to see that the field is populated in that case
                    try:
                        unit = instance.getUnit(field.name)
                        if unit.Filename() != '':
                            value = unit.Filename() #doesn't matter what it is
                    except:
                        pass
                if not value and isinstance(value, FileUpload) and value.filename!='':
                    #OK, its a file, is it empty?
                    value.seek(-1, 2)
                    size = value.tell()
                    value.seek(0)
                    if size == 0:
                        value = None
                    
                if not value:
                    errors[name] =  "%s is required, please correct" % capitalize(name)
                    error = 1
                    break

            #VOCABILARY CHECKS
            if error == 0  and field.enforceVocabulary == 1:
                if value: ## we need to check this as optional field will be
                          ## empty and thats ok
                    values = field.multiValued == 1  and value or [value]
                    for value in values:
                        if not value in field.Vocabulary(instance):
                            errors[name] = "Value %s is not allowed for vocabulary " \
                                           "of element: %s" %(value, capitalize(name))
                            error = 1
                            break

            #CUSTOM VALIDATORS
            if error == 0:
                try:
                    instance.validate_field(name, value, errors)
                except Exception, E:
                    log_exc()
                    errors[name] = E

    security.declarePublic('searchable')
    def searchable(self):
        """return the names of all the searchable fields"""
        return [f.name for f in self.values() if f.searchable]


    #ILayerContainer
    def initalizeLayers(self, instance):
        # scan each field looking for registered layers
        # optionally call its initalizeInstance method and
        # then the initalizeField method
        initalizedLayers = []
        called = lambda x: x in initalizedLayers
        
        for field in self.fields():
            layers = field.registeredLayers()
            for layer, object in layers:
                if not called(layer) and hasattr(object, 'initalizeInstance'):
                    object.initalizeInstance(instance)
                    initalizedLayers.append(layer)
                    
                if hasattr(object, 'initalizeField'):
                    object.initalizeField(instance, field)
                    

class MetadataFieldList(FieldList):
    def addField(self, field):
        """Strictly enforce the contract that metadata is stored w/o
        markup and make sure each field is marked as such for
        generation and introspcection purposes.
        """
        field.isMetadata = 1
        field.storage = MDS
        if 'm' not in field.generateMode:
            field.generateMode = 'mVc'

        
        FieldList.addField(self, field)
    

class Field:
    __implements__ = ILayerContainer
    
    security  = ClassSecurityInfo()
    security.declareObjectPublic()
    security.setDefaultAccess("allow")

    _properties = {
        'required' : 0,
        'default' : None,
        'vocabulary' : (),
        'enforceVocabulary' : 0,
        'multiValued' : 0,
        'searchable' : 0,
        'isMetadata' : 0,

        'accessor' : None,
        'mutator' : None,
        'mode' : 'rw',

        'read_permission' : CMFCorePermissions.View,
        'write_permission' : CMFCorePermissions.ModifyPortalContent,

        'storage' : AttributeStorage(),
        'form_info' : None,

        'generateMode' : 'veVc',
        'force' : '',
        'type' : None,
        }

    def __init__(self, name, **kwargs):
        self.__dict__.update(self._properties)
        self.__dict__.update(kwargs)
        self.name = name
        self.layers = {}

        self.FormInfo()
        
        self.registerLayer('storage', self.storage)

    def registerLayer(self, name, object):
        self.layers[name] = object

    def registeredLayers(self):
        return self.layers.items()
    
    def __repr__(self):
        return "<Field %s(%s:%s)>" %(self.name, self.type, self.mode)

    def _processForm(self):
        form = self.form_info
        if form._processed == 0:
            # Fill in defaults for missing values
            if not form.description:
                name = getattr(form, 'label') or self.name
                form.description = "Enter valid content for %s" % name
            if form.label is None:
                form.label = capitalize(self.name)
            if form.__class__ == DateTimeInfo:
                form.description = form.description + """\n<br/><b>Format</b>  mm/dd/yyyy<br/>"""
            #Lets not do that again ;)
            form._processed = 1
            
    def FormInfo(self):
        form = self.form_info
        if form is None:
            form = self.form_info = StringInfo()

        self._processForm()
        
        return self.form_info

    def Vocabulary(self, content_instance=None):
        value = self.vocabulary
        if not isinstance(value, DisplayList):
            if content_instance:
                method = getattr(content_instance, self.vocabulary, None)
                if method:
                    value = method()
                    
            # Post process value into a DisplayList, templates will use
            # this interface
            sample = value[:1]
            if isinstance(sample, DisplayList):
                #Do nothing, the bomb is already set up
                pass
            elif type(sample) == type(()):
                #Assume we have ( (value, display), ...)
                #and if not ('', '', '', ...)
                if len(sample) != 2:
                    value = zip(value, value)
                value = DisplayList(value)
            elif len(sample) and type(sample[0]) == type(''):
                value = DisplayList(zip(value, value))
            else:
                log("Unhandled type in Vocab")
                log(value)

        return value 


class ObjectField(Field):
    """Base Class for Field objects that fundamentaly deal with raw
    data. This layer implements the interface to IStorage and other
    Field Types should subclass this to delegate through the storage
    layer. 
    """
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'object',
        'default_content_type' : 'application/octet',
        })

    def get(self, instance, **kwargs):
        try:
            return self.storage.get(self.name, instance, **kwargs)
        except AttributeError: # happens if new Atts are added and not yet stored in the instance
            self.storage.set(self.name,instance,self.default,**kwargs)
            return self.default
        
    def set(self, instance, value, **kwargs):
        self.storage.set(self.name, instance, value, **kwargs)

class MetadataField(ObjectField):
    """Metadata fields have special storage and explictly no markup as
    requirements.
    """
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'metadata',
        'isMetadata' : 1,
        'generateMode' : 'mVc',
        'mode' : 'rw',
        'storage' : MetadataStorage(),
        })


class TextField(ObjectField):
    """Base Class for Field objects that rely on some type of
    transformation"""

    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'text',
        'default' : '',
        'default_content_type' : 'text/plain',
        'allowable_content_types' : ('text/plain',),
        'primary' : 0,
        })
                       
    # This deals with a BaseUnit that will handle CachePackets for transforms
    # Fields of this type can have a default view, which can be overridden by the
    # FormInfo if that exists
    default_view = "view"

    def defaultView(self):
        return self.form_info.defaultView() or self.default_view
    
    def set(self, instance, value, **kwargs):
        # We also need to handle the case where there is a baseUnit
        # for this field containing a valid set of data that would
        # not be reuploaded in a subsequent edit, this is basically
        # migrated from the old BaseObject.set method
        if value == '' or type(value) != StringType:
            unit = None
            try:
                unit = ObjectField.get(self, instance)
            except AttributeError:
                pass
            if unit and unit.Filename() != '':
                if ((isinstance(value, FileUpload) and value.filename != '') or
                    (isinstance(value, FileType) and value.name != '')):
                     #OK, its a file, is it empty?
                     value.seek(-1, 2)
                     size = value.tell()
                     value.seek(0)
                     if size == 0:
                         # This new file has no length, so we keep
                         # the orig
                         return
                elif value == '':
                    return #Empty strings don't overwrite either

        mime_type = kwargs.get('mime_type', 'text/plain')
        bu = BaseUnit(self.name, value, mime_type)
        
        ObjectField.set(self, instance, bu, **kwargs)

        #Invoke the default Transforms, hey, its policy
        #Note that we stash the product of transforms on
        #bu.transforms and BU deals with that
        #tt = getToolByName(self, "transformation_tool")
        #tt.runChains(MUTATION,
        #             bu.getRaw(),
        #             bu.transforms)
        

class DateTimeField(ObjectField):
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'datetime'
        })

    def set(self, instance, value, **kwargs):
        if not isinstance(value, DateTime):
            try:
                value = DateTime(value)
            except:
                value = None
            
        ObjectField.set(self, instance, value, **kwargs)
        
class LinesField(ObjectField):
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'lines',
        'default' : [],
        })

    def set(self, instance, value, **kwargs):
        if type(value) == type(''):
            value =  value.split('\n')
        value = [v.strip() for v in value if v.strip()]
        value = filter(None, value)
        ObjectField.set(self, instance, value, **kwargs)

class IntegerField(ObjectField):
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'integer'
        })

    def set(self, instance, value, **kwargs):
        try:
            value = int(value)
        except TypeError:
            value = None
        
        ObjectField.set(self, instance, value, **kwargs)

class SlotField(ObjectField):
    _properties = Field._properties.copy()
    _properties.update({
        'type' : 'slot',
        'generateMode' : 'v',
        })
    

InitializeClass(FieldList)
InitializeClass(Field)
